import jack
import click
import logging
from pathlib import Path
import sys
import signal

from sdnotify import SystemdNotifier
from jack_connection_manager.ConnectionManager import ConnectionManager

logFormat = "%(asctime)s [%(levelname)-5.5s]: %(message)s"
timeFormat = "%Y-%m-%d %H:%M:%S"
logging.basicConfig(format=logFormat, datefmt=timeFormat)
log = logging.getLogger()


# lists for constructing default config paths
default_config_file_path = Path("jack-connection-manager")
default_config_file_name_options = [
    "connections.yml",
    "connection.yml",
    "jack-connection-manager_conf.yml",
    "jack-connection-manager-conf.yml",
    "jack-connection-manager_config.yml",
    "jack-connection-manager-config.yml",
    "config.yml",
    "conf.yml",
]
default_config_file_locations = [
    Path.home() / ".config",
    Path("/etc"),
    Path("/usr/local/etc"),
]


def get_default_config_path():
    for possible_config_path in (
        base / default_config_file_path / filename
        for base in default_config_file_locations
        for filename in default_config_file_name_options
    ):
        if possible_config_path.exists():
            return possible_config_path
    return None


def remove_connections(exclude: list[str]):
    c = jack.Client("jack_disconneeect", no_start_server=True)
    log.info(f"Removing all connections, skipping clients: {', '.join(exclude)}")

    out_ports: list[jack.Port] = c.get_ports(is_audio=True, is_output=True)
    for p in out_ports:
        connected_ports = c.get_all_connections(p)
        if p.name.startswith(exclude):
            continue
        for c_p in connected_ports:
            if c_p.name.startswith(exclude):
                continue

            c.disconnect(p, c_p)


@click.command(
    help="Set Jack Connections",
    context_settings=dict(help_option_names=["-h", "--help"]),
)
@click.option(
    "-c",
    "--config",
    "config_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=Path),
    default=None,
    help="path to connection file",
)
@click.option(
    "-r",
    "--remove-connections",
    "disconnect",
    is_flag=True,
    default=False,
    help="remove all connections except for those specified with -e",
)
@click.option(
    "-e",
    "--exclude",
    multiple=True,
    default=[],
    help="clients that start with the specified string will not be disconnected, can be specified multiple times",
)
@click.option(
    "--client-name", help="Name for the jack client", default="jack_connection_manager"
)
@click.option(
    "-l",
    "--list-missing",
    help="List missing ports and connections",
    is_flag=True,
    default=False,
)
@click.option("-v", "--verbose", count=True, help="increase verbosity level.")
@click.version_option()
def main(config_path, disconnect, exclude, client_name, list_missing, verbose):
    if list_missing:
        log.setLevel(logging.WARN)
    elif verbose == 0:
        log.setLevel(logging.INFO)
    elif verbose >= 1:
        log.setLevel(logging.DEBUG)

    # do the disconnect!
    if disconnect:
        remove_connections(exclude)
        log.info("exiting...")
        sys.exit(0)

    # find config path
    if config_path is None:
        # check different paths for a config file, with the highest one taking precedence
        config_path = get_default_config_path()

    if config_path is None:
        log.error("could not find connection file, please supply one using -c")
        sys.exit(-1)

    cm = ConnectionManager(config_path, client_name)
    if list_missing:
        cm.print_missing_connections()
        sys.exit(0)
    systemd = SystemdNotifier()

    log.info("jack-connection-manager is running")
    for sig in [signal.SIGINT, signal.SIGTERM]:
        signal.signal(sig, cm.deactivate)
    systemd.notify("READY=1")
    # start the connection loop
    cm.connection_loop()


if __name__ == "__main__":

    main()



import yaml
import jack

from pathlib import Path
from time import sleep

import logging
import sys
import queue
from threading import Event, Timer

log = logging.getLogger()

reconnect_wait_time = 2
reconnect_number_retries = 20

retry_timer = 1
n_retries = 15


def add_to_dict_of_sets(d: dict, key, value):
    if key in d:
        d[key].add(value)
    else:
        d[key] = set((value,))


class ConnectionManager:
    def __init__(self, config_path: Path, clientname="jack_connection_manager") -> None:
        # TODO handle jack server not existing

        self.source_ports: dict[str, set[str]] = {}
        self.all_ports: dict[str, set[str]] = {}

        self.build_connection_dict(config_path)

        self.connect_to_jack_server(clientname)
        self.queue: queue.Queue[tuple[jack.Port, jack.Port, int]] = queue.Queue()
        self.stop_event = Event()
        self.c.set_port_registration_callback(self.set_connection_for_port, False)
        self.c.activate()
        self.set_initial_connections()

    def connect_to_jack_server(self, clientname, servername=None):
        n_tries = 0
        while n_tries < reconnect_number_retries:
            try:
                self.c = jack.Client(
                    clientname, no_start_server=True, servername=servername
                )
                break
            except jack.JackOpenError:
                logging.warning("couldn't connect to jack server. retrying...")
                n_tries += 1
                sleep(reconnect_wait_time)
        else:
            logging.error("could not connect to jack server")
            sys.exit(-2)

    def build_connection_dict(self, config_path: Path):
        with open(config_path) as f:
            conf = yaml.load(f, yaml.Loader)

        for source_client in conf:
            source_base = source_client["client"]
            n_channels = source_client["n_channels"]
            source_start_index = (
                source_client["start_index"] if "start_index" in source_client else 1
            )

            for sink_client in source_client["connections"]:
                sink_base = sink_client["client"]
                sink_start_index = (
                    sink_client["start_index"] if "start_index" in sink_client else 1
                )

                for i in range(n_channels):
                    source_port = f"{source_base}{source_start_index+i}"
                    sink_port = f"{sink_base}{sink_start_index+i}"
                    log.debug(f"parsing connection {source_port} -> {sink_port}")
                    add_to_dict_of_sets(self.source_ports, source_port, sink_port)
                    add_to_dict_of_sets(self.all_ports, sink_port, source_port)
                    add_to_dict_of_sets(self.all_ports, source_port, sink_port)

    def set_initial_connections(self):
        for source in self.source_ports:
            try:
                source_port = self.c.get_port_by_name(source)
            except jack.JackError:
                continue

            self.set_connection_for_port(source_port)

    def set_connection_for_port(self, port: jack.Port, registered: bool = True):
        if not registered:
            return

        if not port.name in self.all_ports:
            return

        sinks = self.all_ports[port.name]
        connections = self.c.get_all_connections(port)

        for sink in sinks:
            try:
                sink_port = self.c.get_port_by_name(sink)
            except jack.JackError:
                continue

            if sink_port not in connections:
                log.debug(f"connecting {port.name} -> {sink_port.name}")
                if port.is_output:
                    self.queue.put((port, sink_port, n_retries))
                else:
                    self.queue.put((sink_port, port, n_retries))

    def connection_loop(self):
        """main loop that checks if new connections were put into the conection queue"""
        while not self.stop_event.is_set():
            try:
                out_port, in_port, retries_remaining = self.queue.get(timeout=1)
            except queue.Empty:
                continue
            except TypeError:
                log.error("TypeError while unpacking ports from queue...")
                continue

            try:
                self.c.connect(out_port, in_port)
            except jack.JackErrorCode as e:
                # handle connection already existing
                if e.code == 17:
                    pass
                else:
                    if retries_remaining > 0:
                        log.warning(
                            f"Jack-Error {e.code} while setting connection: {e.message}, retrying..."
                        )
                        t = Timer(
                            retry_timer,
                            self.queue.put,
                            args=((out_port, in_port, retries_remaining - 1),),
                        )
                        t.start()
                    else:
                        log.error(
                            f"Jack-Error {e.code} while setting connection: {e.message}"
                        )

    def print_missing_connections(self):
        missing_ports = set()
        missing_connections = set()
        for source in self.source_ports:
            try:
                source_port = self.c.get_port_by_name(source)
            except jack.JackError:
                print(f"missing port: {source}")
                missing_ports.add(source)
                continue

            sinks = self.source_ports[source]
            connections = self.c.get_all_connections(source_port)

            for sink in sinks:
                try:
                    sink_port = self.c.get_port_by_name(sink)
                except jack.JackError:
                    print(f"missing port: {sink}")
                    missing_ports.add(sink)
                    continue

                if sink_port not in connections:
                    print(f"missing connection {source_port.name} -> {sink_port.name}")
                    missing_connections.add((source, sink))
        return missing_ports, missing_connections

    def deactivate(self, *args):
        log.info("received deactivation signal")
        self.stop_event.set()
        self.c.deactivate()
